/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "controllers/common/family/Families.h"
#include "CyclopsDataHeader.h"
#include <InstrumentManager/InstrumentManager.h>
#include "controllers/common/acquisition/Count.h"
using namespace std;

namespace cyclops {

const string CyclopsDataHeader::TYPE = "cyclops_dataheader";

const string CyclopsDataHeader::SEP_VALUE = ":";
const string CyclopsDataHeader::SEP_PARAM = ",";
const string CyclopsDataHeader::END = "*";

/*
 * Constructor
 */
CyclopsDataHeader::CyclopsDataHeader(const string& name) :
						ExperimentController(name) {

	setFamily(family::HIDDEN);

	registerFunction(TYPE);

	m_CamDriver.init(this, "cammeraDriver");
	m_Title.init(this, "title");
	m_CountSpy.init(this, "count_spy");
	m_SampleSettings.init(this, "sample_settings");
	m_ScanInfo.init(this, "scan_info");
	m_CyclopsSetting.init(this, "cyclops_setting");

	map<string, ExperimentController*> experimentControllersMap =
			InstrumentManager::getInstance()->getAllInstalledExperimentControllersMap();
	map<string, ExperimentController*>::iterator it;
	for (it = experimentControllersMap.begin(); it != experimentControllersMap.end(); it++) {
		if (dynamic_cast<axis::AxisController*>(it->second) != 0) {
			axis::AxisController* axis = dynamic_cast<axis::AxisController*>(it->second);
			m_AxisControllers.push_back(axis);
		}
	}

}

/*
 * Destructor
 */
CyclopsDataHeader::~CyclopsDataHeader() {
}

/*
 * postConfiguration
 */
void CyclopsDataHeader::postConfiguration() {
	registerUpdater(m_CountSpy->statusMessage, &CyclopsDataHeader::updateStatusMessage, this);
}

/*
 * updateProperty
 */
void CyclopsDataHeader::updateStatusMessage() {

	if (m_CountSpy->statusMessage() == acquisition::Count::WRITING_PARAMETERS_MESSAGE) {
		createHeader();
	}
}

/*
 * createHeader
 */
void CyclopsDataHeader::createHeader() {
	ostringstream header;
	header << " ";
	create_a(header, "instrument", m_Title->instrumentName());
	create_a(header, "dated", m_Title->startTime());
	create_a(header, "user", m_Title->userName());
	create_a(header, "l.c.", m_Title->localContactName());
	create_a(header, "commentary", m_Title->title());
	if (m_ScanInfo->running()) {
		create_a(header, "scan_angle", m_ScanInfo->firstControllerName());
	}
	else {
		create_a(header, "scan_angle", "none");
	}
	if (m_ScanInfo->running()) {

		create_i(header, "kctrl", 4, 0);
	}else {
		create_i(header, "kctrl", 5, 0);
	}
	create_i(header, "manip", m_ScanInfo->firstIndex(), 0);
	create_i(header, "nbang", m_ScanInfo->nbController(), 0);
	if (m_ScanInfo->running()) {
		create_i(header, "nkmes", m_ScanInfo->nbStep(), 0);
		create_i(header, "npmes", m_ScanInfo->actualStep(), 0);
	}else {
		create_i(header, "nkmes", 1, 0);
		create_i(header, "npmes", 1, 0);
	}
	create_i(header, "jcode", m_CountSpy->modeint(), 0);
	create_f(header, "count", m_CountSpy->preset()/1000,8 ,2);
	create_i(header, "gain", m_CyclopsSetting->gain(), 0);
	create_i(header, "mode", m_CyclopsSetting->mode(), 0);
	create_i(header, "nacq", m_CyclopsSetting->nbKine(), 0);

	vector<axis::AxisController*>::iterator iter = m_AxisControllers.begin();
	while(iter != m_AxisControllers.end()) {
		ostringstream name;
		name << "angle_" << (*iter)->getName();
		create_f(header, name.str(), (*iter)->position(), 8 ,2);
		++iter;
	}
	if (m_ScanInfo->running()) {
		create_f(header, "scan_start", m_ScanInfo->firstStartValue(), 8 ,2);
		create_f(header, "scan_width", m_ScanInfo->firstWidthValue(), 8 ,2);
		create_f(header, "scan_step", m_ScanInfo->firstStepValue(), 8 ,2);
	}
	else {
		create_f(header, "scan_start", m_ScanInfo->firstStartValue(), 8 ,2);
		create_f(header, "scan_width", 0,8 ,2);
		create_f(header, "scan_step",  0,8  ,2);
	}

	if (m_SampleSettings->useMag() == true) {
		create_f(header, "magf", m_SampleSettings->getMagController()->field(), 8 ,2);
	}
	if (m_SampleSettings->useTemp() == true) {
		create_f(header, "temp", m_SampleSettings->getMainSampleTempController()->regulationTemperature(), 8 ,2);
		create_f(header, "stpt", m_SampleSettings->getMainSampleTempController()->setpointTemperature(), 8 ,2);
		create_f(header, "reqt", m_SampleSettings->getMainSampleTempController()->wantedTemperature(), 8 ,2);
		create_f(header, "samp", m_SampleSettings->getMainSampleTempController()->sampleTemperature(), 8 ,2);
		create_f(header, "heat", m_SampleSettings->getMainSampleTempController()->power(), 8 ,2);
		if (m_SampleSettings->getSecSampleTempController() != NULL) {
			if (m_SampleSettings->getSecSampleTempController()->isEnabled() == true) {
				create_f(header, "temp2", m_SampleSettings->getSecSampleTempController()->regulationTemperature(), 8 ,2);
				create_f(header, "stpt2", m_SampleSettings->getSecSampleTempController()->setpointTemperature(), 8 ,2);
				create_f(header, "reqt2", m_SampleSettings->getSecSampleTempController()->wantedTemperature(), 8 ,2);
				create_f(header, "samp2", m_SampleSettings->getSecSampleTempController()->sampleTemperature(), 8 ,2);
				create_f(header, "heat2", m_SampleSettings->getSecSampleTempController()->power(), 8 ,2);
			}
			else {
				create_f(header, "temp2", -1., 8 ,2);
			}
		}
		else {
			create_f(header, "temp2", -1., 8 ,2);
		}
	}
	header << END;
	m_CamDriver->header = header.str();
}

/*
 * create_a
 */
void CyclopsDataHeader::create_a(ostream& out, const string& name, const string& value) {
	out << name << SEP_VALUE << value  << SEP_PARAM;
}

/*
 * create_i
 */
void CyclopsDataHeader::create_i(ostream& out, const string& name, int32 value, uint32 nbch) {
	ostringstream buf;
	if (nbch > 0) {
		buf.width(nbch);
	}
	buf << value;
	create_a(out, name, buf.str());
}

/*
 * create_f
 */
void CyclopsDataHeader::create_f(ostream& out, const string& name, float64 value, uint32 nbch, uint32 nbdec) {
	ostringstream buf;
	if (nbch > 0) {
		buf.width(nbch);
	}
	buf.precision(nbdec);
	buf << value;
	create_a(out, name, buf.str());
}

}
