/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "PhaTuningController.h"
#include "controllers/common/family/Families.h"

#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>

using namespace std;
using namespace boost;

namespace npp {

const string PhaTuningController::TYPE = "pha_tuning_controller";

/*
 * Constructor
 */
PhaTuningController::PhaTuningController(const string& name) :
		ExperimentController(name), controller::Stoppable(this) {

	setFamily(family::ACQUISITION, family::SETTING);

	displayChannel.init(this, NOSAVE, "display_channel");
	signalRiseTime.init(this, NOSAVE, "rise_time");
	signalDecayTime.init(this, NOSAVE, "decay_time");
	trapezoidRiseTime.init(this, NOSAVE, "trapezoid_rise");
	trapezoidFlatTop.init(this, NOSAVE, "trapezoid_flat");
	triggerSmoothing.init(this, NOSAVE, "trigger_smoothing");
	baselineMean.init(this, NOSAVE, "base_mean");
	peakMean.init(this, NOSAVE, "peak_mean");
	nbControllers.init(this, SAVE, "nb_controllers");
	activeController.init(this, SAVE, "active_controller");

	controllerNumberList.init(this, NOSAVE, "controller_nb_list");
	controllerNameList.init(this, NOSAVE, "controller_name_list");

	registerRefresher(nbControllers, &PhaTuningController::refreshNbControllersProperty, this);
	registerRefresher(activeController, &PhaTuningController::refreshActiveControllersProperty, this);

	registerFunction(TYPE);

	m_registeredControllers = 0;
}

/*
 * Destructor
 */
PhaTuningController::~PhaTuningController() {

}

/*
 * postConfiguration
 */
void PhaTuningController::postConfiguration() {

	controllerNumberList.resize(nbControllers());
	controllerNameList.resize(nbControllers());

	for (int32 i=0; i<m_registeredControllers; ++i) {
		controllerNumberList.set(i, i);
		controllerNameList.set(i, m_phaControllers[i]->getName());
	}

	activeController.setEnumeratedValues(controllerNumberList);
	activeController.setEnumeratedLabels(controllerNameList);
}

void PhaTuningController::start() {

	commandProgression = 0;

	try {
		// Send parameters to ADC card
		int32 value = activeController();
		int32 firstChannel = 0;
		int32 lastChannel = 0;
		switch (displayChannel()) {
		case 0 ... 7:
		firstChannel = displayChannel();
			lastChannel = displayChannel() + 1;
			break;
		case 8:
			firstChannel = 0;
			lastChannel = 4;
			break;
		case 9:
			firstChannel = 4;
			lastChannel = 8;
			break;
		}

		for (int32 channel=firstChannel; channel<lastChannel; ++channel) {
			m_phaControllers[value]->displayChannel = channel;
			m_phaControllers[value]->signalRiseTime = signalRiseTime();
			m_phaControllers[value]->signalDecayTime = signalDecayTime();
			m_phaControllers[value]->trapezoidRiseTime = trapezoidRiseTime();
			m_phaControllers[value]->trapezoidFlatTop = trapezoidFlatTop();
			m_phaControllers[value]->triggerSmoothing = triggerSmoothing();
			m_phaControllers[value]->baselineMean = baselineMean();
			m_phaControllers[value]->peakMean = peakMean();
		}
	} catch(CannotSetValue &e) {
		log(Level::s_Error) << "Cannot set values" << endlog;
	}

	commandProgression = 100;
}

void PhaTuningController::stop() {

	commandProgression = 100;
	log(Level::s_Info) << " stopped" << endlog;
}

void PhaTuningController::updateProgression() {

}

void PhaTuningController::refreshNbControllersProperty(int32 value) throw (CannotSetValue) {

	/**
	 * Check that the number of detectors requested is acceptable
	 */
	if (value == nbControllers()) {
		return;
	} else if (value < nbControllers()) {
		cout << "reducing number of detectors not implemented yet\n";
		return;
	}

	m_phaControllers.resize(value);

	/*
	 * Register detectors for configuration
	 */
	for (int32 i = m_registeredControllers; i < value; ++i) {
		string phaControllerName = "pha_controller" + lexical_cast<string>(i + 1);

		/**
		 * Set the new Detector to null
		 */
		m_phaControllers[i].init(this, phaControllerName);
	}

	m_registeredControllers = value;
}

void PhaTuningController::refreshActiveControllersProperty(int32 value) throw (CannotSetValue) {

	// Get parameters from ADC card
	if (m_phaControllers[value].isAssigned()) {
		displayChannel = m_phaControllers[value]->displayChannel();
		signalRiseTime = m_phaControllers[value]->signalRiseTime();
		signalDecayTime = m_phaControllers[value]->signalDecayTime();
		trapezoidRiseTime = m_phaControllers[value]->trapezoidRiseTime();
		trapezoidFlatTop = m_phaControllers[value]->trapezoidFlatTop();
		triggerSmoothing = m_phaControllers[value]->triggerSmoothing();
		baselineMean = m_phaControllers[value]->baselineMean();
		peakMean = m_phaControllers[value]->peakMean();
	}
}

}
